using DynamicHMC

using DynamicHMC:
    # Hamiltonian
    GaussianKE, Hamiltonian, PhasePoint, neg_energy, phasepoint_in,
    rand_phasepoint, leapfrog, move,
    # building blocks
    rand_bool, TurnStatistic, combine_turnstats, Proposal,
    combined_logprob_logweight, combine_proposals,
    DivergenceStatistic, combine_divstats, divergence_statistic, isdivergent,
    get_acceptance_rate, isturning, adjacent_tree, sample_trajectory,
    # stepsize
    InitialStepsizeSearch, find_initial_stepsize,
    # transitions and tuning
    NUTS_transition, NUTS_init, StepsizeTuner, StepsizeCovTuner, tune,
    TunerSequence, bracketed_doubling_tuner

using Test

using ArgCheck: @argcheck
using DataStructures
using Distributions
using DocStringExtensions: SIGNATURES
import ForwardDiff
using LinearAlgebra
using LogDensityProblems:
    logdensity, dimension, ValueGradient, AbstractLogDensityProblem, LogDensityProblems
using MCMCDiagnostics: effective_sample_size, potential_scale_reduction
using Parameters
import Random
using Random: randn, rand
using StatsBase: mean_and_cov, mean_and_std
using StatsFuns: logaddexp
using Statistics: mean, quantile, Statistics
using Suppressor


# general test environment

const RNG = Random.GLOBAL_RNG   # shorthand
Random.seed!(RNG, UInt32[0x23ef614d, 0x8332e05c, 0x3c574111, 0x121aa2f4])

"Tolerant testing in a CI environment."
const RELAX = (k = "CONTINUOUS_INTEGRATION"; haskey(ENV, k) && ENV[k] == "true")

"""
    $SIGNATURES

Random positive definite matrix of size `n` x `n` (for testing).
"""
function rand_Σ(::Type{Symmetric}, n)
    A = randn(n, n)
    Symmetric(A'*A .+ 0.01)
end

rand_Σ(::Type{Diagonal}, n) = Diagonal(randn(n).^2 .+ 0.01)

rand_Σ(n::Int) = rand_Σ(Symmetric, n)


# use multivariate distributions as tests

"""
Obtain the log density from a distribution.
"""
struct DistributionLogDensity{D <: Distribution{Multivariate,Continuous}
                              } <: AbstractLogDensityProblem
    distribution::D
end

LogDensityProblems.dimension(ℓ::DistributionLogDensity) = length(ℓ.distribution)

LogDensityProblems.logdensity(::Type{ValueGradient}, ℓ::DistributionLogDensity, x) =
    ValueGradient(logpdf(ℓ.distribution, x), gradlogpdf(ℓ.distribution, x))

DistributionLogDensity(::Type{MvNormal}, n::Int) = # canonical
    DistributionLogDensity(MvNormal(zeros(n), ones(n)))

Statistics.mean(ℓ::DistributionLogDensity) = mean(ℓ.distribution)
Statistics.var(ℓ::DistributionLogDensity) = var(ℓ.distribution)
Statistics.cov(ℓ::DistributionLogDensity) = cov(ℓ.distribution)
Base.rand(ℓ::DistributionLogDensity) = rand(ℓ.distribution)

"""
A function returning a log density (as a `ValueGradient`).
"""
struct FunctionLogDensity{F} <: AbstractLogDensityProblem
    dimension::Int
    f::F
end

LogDensityProblems.dimension(ℓ::FunctionLogDensity) = length(ℓ.distribution)

LogDensityProblems.logdensity(::Type{ValueGradient}, ℓ::FunctionLogDensity, x) =
    ℓ.f(x)::ValueGradient

"""
$(SIGNATURES)

A log density always returning a constant (for testing).
"""
FunctionLogDensity(v::ValueGradient) = FunctionLogDensity(length(v.gradient), _ -> v)

"Random Hamiltonian `H` with phasepoint `z`, with dimension `K`."
function rand_Hz(K)
    μ = randn(K)
    Σ = rand_Σ(K)
    κ = GaussianKE(inv(rand_Σ(Diagonal, K)))
    H = Hamiltonian(DistributionLogDensity(MvNormal(μ, Matrix(Σ))), κ)
    z = rand_phasepoint(RNG, H, μ)
    H, z
end

include("test-Hamiltonian-leapfrog.jl")
include("test-buildingblocks.jl")
include("test-stepsize.jl")
include("test-sample-dummy.jl")
include("test-tuners.jl")
include("test-sample-normal.jl")
include("test-normal-mcmc.jl")
include("test-statistics.jl")
include("test-reporting.jl")

include("../docs/make.jl")
